;;
;; hackerbase.scm
;;
;; Hackerspace Members Database management system.
;;
;; ISC License
;;
;; Copyright 2023 Brmlab, z.s.
;; Dominik Pantůček <dominik.pantucek@trustica.cz>
;;
;; Permission to use, copy, modify, and/or distribute this software
;; for any purpose with or without fee is hereby granted, provided
;; that the above copyright notice and this permission notice appear
;; in all copies.
;; 
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
;; WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
;; WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
;; AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
;; CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
;; OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
;; NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
;; CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

(import (chicken repl)
	month
	command-line
	mbase
	brmember
	configuration
	cards
	members-print
	members-payments
	web-static
	environment
	mailman
	texts
	tests
	reminders
	util-mail
	logging
	progress
	period)

;; Command-line options and configurable parameters
(define -needs-bank- (make-parameter #f))
(define -member-id- (make-parameter #f))
(define -member-nick- (make-parameter #f))
(define -action- (make-parameter #f))
(define -fname- (make-parameter #f))
(define -run-tests?- (make-parameter #f))
(define -web-dir- (make-parameter #f))
(define -normal-month- (make-parameter #t))
(define -ml-all- (make-parameter #f))
(define -show-destroyed- (make-parameter #f))
(define -notify-months- (make-parameter 1))
(define -send-emails- (make-parameter #f))
(define -quiet- (make-parameter #f))

;; Configuration overrides
(define =members-directory= (make-parameter #f))
(define =apikeys-file= (make-parameter #f))
(define =jendasap-checked= (make-parameter #f))
(define =bank-dir= (make-parameter #f))
(define =email-from= (make-parameter #f))

;; Arguments parsing
(command-line
 print-help
 (-h () "This help"
     (print "Command-line options:")
     (newline)
     (print-help)
     (newline)
     (exit 0))
 (-license () "Show licensing terms"
	   (print banner)
	   (newline)
	   (print license)
	   (exit 0))
 ""
 "Configuration options:"
 (-config (fname) "Initial configuration"
	  (*etc-hackerbase* fname))
 (-members (dir) "Members base directory"
	   (=members-directory= dir))
 (-month (YYYY-MM) "Specify current month"
	 (-normal-month- #f)
	 (*current-month* (string->month YYYY-MM)))
 (-tstyle (style) "Use given table style: debug, ascii, unicode"
	  (*table-border-style* (string->symbol style)))
 (-apikey (fname) "File with Fio API keys"
	  (=apikeys-file= fname))
 (-bankdir (dir) "Where are bank CSV files"
	   (=bank-dir= dir))
 (-checked (file) "JendaSAP checked.ntlm file"
	   (=jendasap-checked= file))
 (-ml-all () "Load all mailman lists"
	  (-ml-all- #t))
 (-destroyed () "Show destroyed members in -fees"
	     (-show-destroyed- #t))
 (-sendmail () "Actually send emails"
	    (-send-emails- #t))
 (-mailto (email) "Override all outgoing emails destination"
	  (*mailto-override* email))
 (-logfile (filename) "Enable logging to file"
	   (*log-file* filename))
 (-quiet () "Suppress all output"
	 (-quiet- #t)
	 (*progress-quiet* #t))
 (-from (email) "Sender email address"
	(=email-from= email))
 ""
 "Query options:"
 (-mi (id) "Specify member by id" (-member-id- (string->number id)))
 (-mn (nick) "Specify member by nick"
      (when (not (-action-))
	(-action- 'no-op))
      (-member-nick- nick))
 ""
 "Actions:"
 (-info () "Print information"
	(-needs-bank- #t)
	(-action- 'print-info))
 (-stats (file:gnuplot-data) "Get stats for all months"
	 (-action- 'print-stats)
	 (-fname- file:gnuplot-data))
 (-print () "Print given member file"
	 (-action- 'print-member-file))
 (-tests () "Run self-tests upon startup" (-run-tests?- #t))
 (-idstats () "Returns information about available member ids" (-action- 'print-idstats))
 (-genid () "Generates random member id" (-action- 'genid))
 (-gencards (file:cards file:desfires) "Generates brmdoor-compatible card files"
	    (-action- 'gencards)
	    (-fname- (list file:cards file:desfires)))
 (-problems () "Prints all files with problems" (-action- 'problems))
 (-fees () "Prints fees table"
	(-needs-bank- #t)
	(-action- 'fees))
 (-genweb (dir) "Generate static web files"
	  (-needs-bank- #t)
	  (-web-dir- dir)
	  (-action- 'genweb))
 (-repl () "Start REPL with everything loaded"
	(-needs-bank- #t)
	(-action- 'repl))
 (-edit () "Edit selected user"
	(-action- 'edit))
 (-unpaired () "Show latest unpaired bank transactions"
	    (-needs-bank- #t)
	    (-action- 'unpaired))
 (-mlcheck () "Check internal ML"
	   (-action- 'mlcheck))
 (-mlsync () "Synchronize internal ML"
	  (-action- 'mlsync))
 (-notify () "Members with debt for more than 1 month"
	  (-notify-months- 1)
	  (-needs-bank- #t)
	  (-action- 'notify))
 (-notify3 () "Members with debt for more than 3 month"
	   (-notify-months- 3)
	   (-needs-bank- #t)
	   (-action- 'notify))
 )

;; Print banner
(when (not (-quiet-))
  (print banner-line)
  (newline))

;; Load default configuration
(load-configuration!)

;; Override from command-line
(when (=members-directory=)
  (*members-directory* (=members-directory=)))
(when (=apikeys-file=)
  (*apikeys-file* (=apikeys-file=)))
(when (=jendasap-checked=)
  (*jendasap-checked* (=jendasap-checked=)))
(when (=bank-dir=)
  (*bank-dir* (=bank-dir=)))
(when (=email-from=)
  (*email-from* (=email-from=)))

;; Run tests
(when (-run-tests?-)
  (run-all-tests!)
  (newline))

;; Start real work
(log-info "Loading")

;; Load the members database (required for everything anyway)
(define MB0
  (if (-action-)
      (let ((mb (load-mbase (*members-directory*) #t)))
	(if (-needs-bank-)
	    (members-payments-process mb (*apikeys-file*) (*bank-dir*) (*jendasap-checked*))
	    mb))
      #f))

;; Load ML(s) and merge them
(define-values (MB internal-ml)
  (if MB0
      (if (-ml-all-)
	  (let ()
	    (define mls (load-mailman-lists))
	    (values (foldl (lambda (mb ml)
			     (mbase-merge-mailman mb ml))
			   MB0
			   mls)
		    (find-mailman-list mls "internal")))
	  (let ()
	    (define internal-ml (load-mailman-list "internal"))
	    (values (mbase-merge-mailman MB0 internal-ml)
		    internal-ml)))
      (values #f #f)))

;; If a member is specified by either id or nick, get its record
(define mr
  (if (-member-id-)
      (let ((mr (find-member-by-id MB (-member-id-))))
	(when (not mr)
	  (print "Member id " (-member-id-) " not found!"))
	mr)
      (if (-member-nick-)
	  (let ((mrs (find-members-by-nick MB (-member-nick-))))
	    (cond ((null? mrs)
		   (newline)
		   (print "Member nick " (-member-nick-) " not found!")
		   #f)
		  ((> (length mrs) 1)
		   (newline)
		   (print "Found: " (member-records->string mrs))
		   (-action- 'no-op)
		   #f)
		  (else
		   (car mrs))))
	  #f)))

;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Perform requested action
(case (-action-)
  ((print-info)
   (newline)
   (print "Current month: " (month->string (*current-month*)))
   (newline)
   (if mr
       (print-member-table mr)
       (print-members-base-table MB))
   (newline))
  ((print-stats)
   (newline)
   (parameterize ((current-output-port (open-output-file (-fname-))))
     (print-members-base-stats
      (mbase-stats MB))))
  ((print-member-file)
   (cond (mr
	  (newline)
	  (print-member-source mr))
	 (else
	  (print "No member specified!"))))
  ((print-idstats)
   (newline)
   (print-members-ids-stats MB)
   (newline))
  ((genid)
   (newline)
   (print "New member id: " (mbase-gen-id MB))
   (newline))
  ((gencards)
   (if (-normal-month-)
       (apply cards-export MB (-fname-))
       (print "Cards export disabled with manually specified current month.")))
  ((problems)
   (let ((num (let loop ((mb (mbase-members MB))
			 (num 0))
		(if (null? mb)
		    num
		    (loop (cdr mb)
			  (if (or (brmember-has-problems? (car mb))
				  (brmember-has-highlights? (car mb)))
			      (let ()
				(newline)
				(print-member-table (car mb))
				(print-member-source (car mb))
				(add1 num))
			      num))))))
     (when (= num 0)
       (newline)
       (print "No problems found."))))
  ((fees)
   (newline)
   (print-members-fees-table MB (-show-destroyed-)))
  ((repl)
   (repl))
  ((genweb)
   (log-info "Generating static web files")
   (gen-web-static MB (-web-dir-)))
  ((edit)
   (if mr
       (edit-file (brmember-file-path mr))
       (print "No member to edit.")))
  ((no-op) (void))
  ((unpaired)
   (newline)
   (print-unpaired-table MB))
  ((mlsync)
   (if (-normal-month-)
       (mailman-sync-members internal-ml (mbase-active-emails MB))
       (print "Mailman synchronization disabled with manually specified current month.")))
  ((mlcheck)
   (let-values (((missing surplus)
		 (mailman-compare-members internal-ml
					  (mbase-active-emails MB))))
     (if (and (null? missing)
	      (null? surplus))
	 (print "Internal mailing list membership in sync.")
	 (let ()
	   (print "Internal mailing list:")
	   (when (not (null? missing))
	     (print "  Missing: " missing))
	   (when (not (null? surplus))
	     (print "  Outsiders: " surplus))))))
  ((notify)
   (let ((nmembers (members-to-notify MB (-notify-months-))))
     (newline)
     (if (null? nmembers)
	 (print "Everyone paid on time.")
	 (let ()
	   (print "Notify" (-notify-months-))
	   (let loop ((lst nmembers))
	     (when (and (not (null? lst))
			(or (not mr)
			    (eq? (brmember-id mr)
				 (brmember-id (car lst)))))
	       (if (-send-emails-)
		   (make+send-reminder-email (car lst))
		   (make+print-reminder-email (car lst)))
	       (loop (cdr lst))))))))
  (else
   (print "Nothing to do."))
  
  )

;; Finished
(log-info "Finished")
