;;
;; util-io.scm
;;
;; Special IO extensions to deal with weird stuff.
;;
;; ISC License
;;
;; Copyright 2023 Brmlab, z.s.
;; Dominik Pantůček <dominik.pantucek@trustica.cz>
;;
;; Permission to use, copy, modify, and/or distribute this software
;; for any purpose with or without fee is hereby granted, provided
;; that the above copyright notice and this permission notice appear
;; in all copies.
;; 
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
;; WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
;; WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
;; AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
;; CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
;; OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
;; NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
;; CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

(declare (unit util-io))

(module
 util-io
 (
  read-lines/no-bom
  get-process-output-lines
  process-send/recv
  )

 (import scheme
	 (chicken base)
	 (chicken io)
	 (chicken process))

 ;; If given string begins with UTF-8 BOM, it is removed.
 (define (remove-optional-bom str)
   (if (< (string-length str) 3)
       str
       (let ((maybe-bom (substring str 0 3)))
	 (if (string=? maybe-bom "\xEF\xBB\xBF")
	     (substring str 3)
	     str))))

 ;; Reads lines from given input port, discarding BOM at the beginning
 ;; of the first line if there is any.
 (define (read-lines/no-bom ip)
   (let ((lines (read-lines ip)))
     (if (null? lines)
	 lines
	 (cons (remove-optional-bom (car lines))
	       (cdr lines)))))

 ;; Very simple shell command wrapper that returns lines produced by
 ;; given command. Dangerous - performs no argument escaping!
 (define (get-process-output-lines cmd . args)
   (let-values (((stdout stdin pid stderr) (process* cmd args)))
     (close-output-port stdin)
     (let ((result (read-lines stdout)))
       (let-values (((a b c) (process-wait pid)))
	 result))))

 ;; Invokes given command with given arguments, gives it all input
 ;; lines and returns the output lines.
 (define (process-send/recv cmd args . lines)
   (let-values (((stdout stdin pid stderr) (process* cmd args)))
     (let loop ((lines lines))
       (when (not (null? lines))
	 (display (car lines) stdin)
	 (newline stdin)
	 (loop (cdr lines))))
     (close-output-port stdin)
     (let ((result (read-lines stdout)))
       (let-values (((pid exit-ok? exit/signal) (process-wait pid)))
	 result))))

 )
